# Refactoring base Lib from TFDS

"""
adapted from https://github.com/tensorflow/datasets/blob/v4.4.0/tensorflow_datasets/core/utils/gcs_utils.py
"""


"""Utilities for accessing TFDS GCS buckets."""

import concurrent.futures
import os
import posixpath
from typing import List, Optional


from ..utils import lazy_import

tf = lazy_import('tensorflow.compat.v2')

from fileio.src import type_utils
from fileio.src import generic_path
from fileio.src import py_utils


GCS_ROOT_DIR = 'gs://tfds-data'

# for dataset_info/
GCS_DATASET_INFO_DIR = 'dataset_info'
GCS_DATASETS_DIR = 'datasets'

_is_gcs_disabled = False

# Exception raised when GCS isn't available
# * UnimplementedError: On windows, gs:// isn't supported on old TF versions.
#     https://github.com/tensorflow/tensorflow/issues/38477
# * FailedPreconditionError: (e.g. no internet)
# * PermissionDeniedError: Some environments block GCS access.
# * AbortedError: All 10 retry attempts failed.
GCS_UNAVAILABLE_EXCEPTIONS = (
    tf.errors.UnimplementedError,
    tf.errors.FailedPreconditionError,
    tf.errors.PermissionDeniedError,
    tf.errors.AbortedError,
)


def gcs_path(*relative_path: type_utils.PathLike, gcs_bucket=None, overwrite_global=False) -> type_utils.ReadWritePath:
    """Returns the GCS URI path.
    Args:
        *relative_path: Eventual relative path in the bucket.
    Returns:
        path: The GCS uri.
    """
    global GCS_ROOT_DIR
    bucket_path = GCS_ROOT_DIR or gcs_bucket
    if gcs_bucket and overwrite_global:
        GCS_ROOT_DIR = generic_path.as_path(gcs_bucket).parent
    return generic_path.as_path(bucket_path).joinpath(*relative_path)


# Community datasets index.
# This file contains the list of all community datasets with their associated
# location.
# Datasets there are downloaded and installed locally by the
# `PackageRegister` during `tfds.builder`
GCS_COMMUNITY_INDEX_PATH = gcs_path() / 'community-datasets-list.jsonl'


def exists(path: type_utils.ReadWritePath) -> bool:
    """Checks if path exists. Returns False if issues occur connecting to GCS."""
    try:
        return path.exists()
    except GCS_UNAVAILABLE_EXCEPTIONS:
        return False


@py_utils.memoize()
def gcs_listdir(dir_name: str) -> Optional[List[str]]:
    """List all files in the given GCS dir (`['dataset/1.0.0/file0', ...]`)."""
    root_dir = gcs_path(dir_name)
    if _is_gcs_disabled or not exists(root_dir):
        return None
    return [posixpath.join(dir_name, f.name) for f in root_dir.iterdir()]



"""
adapted from https://github.com/tensorflow/datasets/blob/master/tensorflow_datasets/core/file_adapters.py
"""

"""Adapters for file formats."""

import abc
import enum
import os

from typing import Any, ClassVar, Dict, Iterable, List, Optional, Type

from ..utils import lazy_import
tf = lazy_import('tensorflow')
from fileio.src import type_utils

ExamplePositions = List[Any]


class FileFormat(enum.Enum):
    """Format of the record files.
    The values of the enumeration are used as filename endings/suffix.
    """
    TFRECORD = 'tfrecord'
    RIEGELI = 'riegeli'


DEFAULT_FILE_FORMAT = FileFormat.TFRECORD


class FileAdapter(abc.ABC):
    """Interface for Adapter objects which read and write examples in a format."""

    FILE_SUFFIX: ClassVar[str]

    @classmethod
    @abc.abstractmethod
    def make_tf_data(
            cls,
            filename: type_utils.PathLike,
            buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        """Returns TensorFlow Dataset comprising given record file."""
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def write_examples(
            cls,
            path: type_utils.PathLike,
            iterator: Iterable[type_utils.KeySerializedExample],
    ) -> Optional[ExamplePositions]:
        """Write examples from given iterator in given path.
        Args:
            path: Path where to write the examples.
            iterator: Iterable of examples.
        Returns:
            List of record positions for each record in the given iterator. In case of
            TFRecords, does not return anything.
        """
        raise NotImplementedError()


class TfRecordFileAdapter(FileAdapter):
    """File adapter for TFRecord file format."""

    FILE_SUFFIX = 'tfrecord'

    @classmethod
    def make_tf_data(
            cls,
            filename: type_utils.PathLike,
            buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        """Returns TensorFlow Dataset comprising given record file."""
        return tf.data.TFRecordDataset(filename, buffer_size=buffer_size)

    @classmethod
    def write_examples(
            cls,
            path: type_utils.PathLike,
            iterator: Iterable[type_utils.KeySerializedExample],
    ) -> Optional[ExamplePositions]:
        """Write examples from given iterator in given path.
        Args:
            path: Path where to write the examples.
            iterator: Iterable of examples.
        Returns:
            None
        """
        with tf.io.TFRecordWriter(os.fspath(path)) as writer:
            for _, serialized_example in iterator:
                writer.write(serialized_example)
            writer.flush()


class RiegeliFileAdapter(FileAdapter):
    """File adapter for Riegeli file format."""

    FILE_SUFFIX = 'riegeli'

    @classmethod
    def make_tf_data(
            cls,
            filename: type_utils.PathLike,
            buffer_size: Optional[int] = None,
    ) -> tf.data.Dataset:
        from riegeli.tensorflow.ops import riegeli_dataset_ops as riegeli_tf    # pylint: disable=g-import-not-at-top
        return riegeli_tf.RiegeliDataset(filename, buffer_size=buffer_size)

    @classmethod
    def write_examples(
            cls,
            path: type_utils.PathLike,
            iterator: Iterable[type_utils.KeySerializedExample],
    ) -> Optional[ExamplePositions]:
        """Write examples from given iterator in given path.
        Args:
            path: Path where to write the examples.
            iterator: Iterable of examples.
        Returns:
            List of record positions for each record in the given iterator.
        """
        positions = []
        import riegeli    # pylint: disable=g-import-not-at-top
        with tf.io.gfile.GFile(os.fspath(path), 'wb') as f:
            with riegeli.RecordWriter(f, options='transpose') as writer:
                for _, record in iterator:
                    writer.write_record(record)
                    positions.append(writer.last_pos)
        return positions


def _to_bytes(key: type_utils.Key) -> bytes:
    """Convert the key to bytes."""
    if isinstance(key, int):
        return key.to_bytes(128, byteorder='big')    # Use 128 as this match md5
    elif isinstance(key, bytes):
        return key
    elif isinstance(key, str):
        return key.encode('utf-8')
    else:
        raise TypeError(f'Invalid key type: {type(key)}')


# Create a mapping from FileFormat -> FileAdapter.
ADAPTER_FOR_FORMAT: Dict[FileFormat, Type[FileAdapter]] = {
    FileFormat.RIEGELI: RiegeliFileAdapter,
    FileFormat.TFRECORD: TfRecordFileAdapter,
}

_FILE_SUFFIX_TO_FORMAT = {
    adapter.FILE_SUFFIX: file_format
    for file_format, adapter in ADAPTER_FOR_FORMAT.items()
}


def file_format_from_suffix(file_suffix: str) -> FileFormat:
    """Returns the file format associated with the file extension (`tfrecord`)."""
    if file_suffix not in _FILE_SUFFIX_TO_FORMAT:
        raise ValueError('Unrecognized file extension: Should be one of '
                                         f'{list(_FILE_SUFFIX_TO_FORMAT.values())}')
    return _FILE_SUFFIX_TO_FORMAT[file_suffix]


def is_example_file(filename: str) -> bool:
    """Whether the given filename is a record file."""
    return any(f'.{adapter.FILE_SUFFIX}' in filename
                         for adapter in ADAPTER_FOR_FORMAT.values())


from tqdm.auto import tqdm
from .core import get_pathlike, json, jparser

# improved JSON parser

def read_jsonlines(filepath, ignore_errors: bool = True, parse_obj: bool = False, **kwargs):
    filepath = get_pathlike(filepath)
    with filepath.open(mode='rb+') as f:
        pbar = tqdm(f, desc=filepath.name)
        for _, line in enumerate(pbar):
            try:
                l = jparser.parse(line, recursive=parse_obj, **kwargs)
                yield l
                del l

            except Exception as e:
                if ignore_errors:
                    continue
                raise e
            pbar.update()
        pbar.close()


def read_json(filepath, parse_obj: bool = False, **kwargs):
    filepath = get_pathlike(filepath)
    if not parse_obj:
        return json.load(filepath.read_bytes(), **kwargs)
    return jparser.parse(filepath.read_bytes(), **kwargs)


def autojson(filepath, parse_obj: bool = False, ignore_errors: bool = True, **kwargs):
    dtype = filepath.file_ext
    if dtype == 'json':
        return read_json(filepath, parse_obj=parse_obj, **kwargs)
    return read_jsonlines(filepath, ignore_errors=ignore_errors, parse_obj=parse_obj, **kwargs)







from . import utils
from . import src

from .src import (
    gfile,
    glob,
    gcopy,
    isdir,
    isfile,
    listdir,
    mkdirs,
    mv,
    exists,
    rmdir,
    rm, 
    jparser,
    json,
    TextLineDataset,
    TFRecordDataset,
    TFRecordWriter,
    timestamp,
    ftimestamp,
    lazy_install,
    lazy_import

)
from .src import File

from .utils import MultiThreadPipeline, TFDSIODataset
from .src import LineSeekableFile, iterator_function
from .src import (
    torchdevice,
    device,
    _pickler,
    pickler,
    PathIO,
    PathIOLike,
    as_path,
)

__all__ = [
    "File",
    "gfile",
    "_pickler",
    "pickler",
    "LineSeekableFile",
    "timestamp",
    "ftimestamp",
    "as_path",
    "PathIOLike",
    "PathIO",
    "glob",
    "gcopy",
    "isdir",
    "isfile",
    "listdir",
    "mkdirs",
    "mv",
    "exists",
    "rmdir",
    "rm", 
    "jparser",
    "json",
    "TextLineDataset",
    "TFRecordDataset",
    "TFRecordWriter",
    "lazy_install",
    "lazy_import",
    "iterator_function",
    "device",
    "torchdevice"
]



"""Pathlib-like generic abstraction.
adapted from https://github.com/tensorflow/datasets/blob/v4.4.0/tensorflow_datasets/core/utils/generic_path.py
"""

import os
import typing
from typing import Callable, Dict, Tuple, Type, Union, TypeVar

from fileio.src import gpath
from fileio.src import type_utils

PathLike = type_utils.PathLike
ReadOnlyPath = type_utils.ReadOnlyPath
ReadWritePath = type_utils.ReadWritePath

PathLikeCls = Union[Type[ReadOnlyPath], Type[ReadWritePath]]
T = TypeVar('T')

_PATHLIKE_CLS: Tuple[PathLikeCls, ...] = (
    gpath.PosixGPath,
    gpath.WindowsGPath,
)
_URI_PREFIXES_TO_CLS: Dict[str, PathLikeCls] = {
    # Even on Windows, `gs://`,... are PosixPath
    uri_prefix: gpath.PosixGPath for uri_prefix in gpath.URI_PREFIXES
}


# pylint: disable=g-wrong-blank-lines
@typing.overload
def register_pathlike_cls(path_cls_or_uri_prefix: str) -> Callable[[T], T]:
    ...


@typing.overload
def register_pathlike_cls(path_cls_or_uri_prefix: T) -> T:
    ...


def register_pathlike_cls(path_cls_or_uri_prefix):
    """Register the class to be forwarded as-is in `as_path`.
    ```python
    @utils.register_pathlike_cls('my_path://')
    class MyPath(pathlib.PurePosixPath):
        ...
    my_path = tfds.core.as_path('my_path://some-path')
    ```
    Args:
        path_cls_or_uri_prefix: If a uri prefix is given, then passing calling
            `tfds.core.as_path('prefix://path')` will call the decorated class.
    Returns:
        The decorator or decoratorated class
    """
    global _PATHLIKE_CLS
    if isinstance(path_cls_or_uri_prefix, str):

        def register_pathlike_decorator(cls: T) -> T:
            _URI_PREFIXES_TO_CLS[path_cls_or_uri_prefix] = cls
            return register_pathlike_cls(cls)

        return register_pathlike_decorator
    else:
        _PATHLIKE_CLS = _PATHLIKE_CLS + (path_cls_or_uri_prefix,)
        return path_cls_or_uri_prefix


# pylint: enable=g-wrong-blank-lines


def as_path(path: PathLike) -> ReadWritePath:
    """Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.
    Args:
        path: Pathlike object.
    Returns:
        path: The `pathlib.Path`-like abstraction.
    """
    is_windows = os.name == 'nt'
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    # str is URI (e.g. `gs://`, `github://`,...)
            # On windows, `PosixGPath` is created for `gs://` paths
            return _URI_PREFIXES_TO_CLS[uri_splits[0] + '://'](path)    # pytype: disable=bad-return-type
        elif is_windows:
            return gpath.WindowsGPath(path)
        else:
            return gpath.PosixGPath(path)
    elif isinstance(path, _PATHLIKE_CLS):
        return path    # Forward resource path, gpath,... as-is    # pytype: disable=bad-return-type
    elif isinstance(path, os.PathLike):    # Other `os.fspath` compatible objects
        path_cls = gpath.WindowsGPath if is_windows else gpath.PosixGPath
        return path_cls(path)
    else:
        raise TypeError(f'Invalid path type: {path!r}')





"""Typing annotation utils.
Aadapted from https://github.com/tensorflow/datasets/blob/v4.4.0/tensorflow_datasets/core/utils/type_utils.py
"""


import abc
import os
import typing
from typing import Any, AnyStr, Dict, Iterator, List, Optional, Sequence, Tuple, Type, TypeVar, Union

from ..utils import lazy_import
tf = lazy_import('tensorflow')

# pylint: disable=g-import-not-at-top
try:
    from typing import Protocol
except ImportError:
    import typing_extensions
    Protocol = typing_extensions.Protocol
# pylint: enable=g-import-not-at-top

# Accept both `str` and `pathlib.Path`-like
PathLike = Union[str, os.PathLike]
PathLikeCls = (str, os.PathLike)    # Used in `isinstance`

T = TypeVar('T')

# Note: `TupleOrList` avoid abiguity from `Sequence` (`str` is `Sequence[str]`,
# `bytes` is `Sequence[int]`).
TupleOrList = Union[Tuple[T, ...], List[T]]

TreeDict = Union[T, Dict[str, 'TreeDict']]    # pytype: disable=not-supported-yet
Tree = Union[T, TupleOrList['Tree'], Dict[str, 'Tree']]    # pytype: disable=not-supported-yet

Tensor = Union[tf.Tensor, tf.SparseTensor, tf.RaggedTensor]

Dim = Optional[int]
Shape = TupleOrList[Dim]

JsonValue = Union[str, bool, int, float, None, List['JsonValue'],
                                    Dict[str, 'JsonValue'],    # pytype: disable=not-supported-yet
                                 ]
Json = Dict[str, JsonValue]

# Types for the tfrecord example construction.

Key = Union[int, str, bytes]
KeySerializedExample = Tuple[Key, bytes]    # `(key, serialized_proto)`

# pytype: disable=ignored-abstractmethod


class PurePath(Protocol):
    """Protocol for pathlib.PurePath-like API."""
    parts: Tuple[str, ...]
    drive: str
    root: str
    anchor: str
    name: str
    suffix: str
    suffixes: List[str]
    stem: str

    # pylint: disable=multiple-statements,line-too-long

    def __new__(cls: Type[T], *args: PathLike) -> T:
        raise NotImplementedError

    def __fspath__(self) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        raise NotImplementedError

    def __lt__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __le__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __gt__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __ge__(self, other: 'PurePath') -> bool:
        raise NotImplementedError

    def __truediv__(self: T, key: PathLike) -> T:
        raise NotImplementedError

    def __rtruediv__(self: T, key: PathLike) -> T:
        raise NotImplementedError

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def as_posix(self) -> str:
        raise NotImplementedError

    def as_uri(self) -> str:
        raise NotImplementedError

    def is_absolute(self) -> bool:
        raise NotImplementedError

    def is_reserved(self) -> bool:
        raise NotImplementedError

    def match(self, path_pattern: str) -> bool:
        raise NotImplementedError

    def relative_to(self: T, *other: PathLike) -> T:
        raise NotImplementedError

    def with_name(self: T, name: str) -> T:
        raise NotImplementedError

    def with_suffix(self: T, suffix: str) -> T:
        raise NotImplementedError

    def joinpath(self: T, *other: PathLike) -> T:
        raise NotImplementedError

    @property
    def parents(self: T) -> Sequence[T]:
        raise NotImplementedError

    @property
    def parent(self: T) -> T:
        raise NotImplementedError
    
    @property
    def absolute_parent(self: T) -> T:
        raise NotImplementedError
    
    @property
    def is_cloud(self: T) -> bool:
        raise NotImplementedError

    @property
    def is_gcs(self: T) -> bool:
        raise NotImplementedError

    # py3.9 backport of PurePath.is_relative_to.
    def is_relative_to(self, *other: PathLike) -> bool:
        """Return True if the path is relative to another path or False."""
        try:
            self.relative_to(*other)
            return True
        except ValueError:
            return False

    # pylint: enable=multiple-statements,line-too-long


class ReadOnlyPath(PurePath, Protocol):
    """Protocol for read-only methods of pathlib.Path-like API.
    See [pathlib.Path](https://docs.python.org/3/library/pathlib.html)
    documentation.
    """

    def __new__(cls: Type[T], *args: PathLike) -> T:
        if cls in (ReadOnlyPath, ReadWritePath):
            from tensorflow_datasets.core.utils import generic_path    # pytype: disable=import-error    # pylint: disable=g-import-not-at-top
            return generic_path.as_path(*args)
        else:
            return super().__new__(cls, *args)

    @abc.abstractmethod
    def exists(self) -> bool:
        """Returns True if self exists."""

    @abc.abstractmethod
    def is_dir(self) -> bool:
        """Returns True if self is a dir."""

    def is_file(self) -> bool:
        """Returns True if self is a file."""
        return not self.is_dir()

    @abc.abstractmethod
    def iterdir(self: T) -> Iterator[T]:
        """Iterates over the directory."""

    @abc.abstractmethod
    def glob(self: T, pattern: str) -> Iterator[T]:
        """Yielding all matching files (of any kind)."""
        # Might be able to implement using `iterdir` (recursivelly for `rglob`).

    def rglob(self: T, pattern: str) -> Iterator[T]:
        """Yielding all matching files recursivelly (of any kind)."""
        return self.glob(f'**/{pattern}')

    def expanduser(self: T) -> T:
        """Returns a new path with expanded `~` and `~user` constructs."""
        if '~' not in self.parts:    # pytype: disable=attribute-error
            return self
        raise NotImplementedError

    @abc.abstractmethod
    def resolve(self: T, strict: bool = False) -> T:
        """Returns the absolute path."""

    @abc.abstractmethod
    def open(
            self,
            mode: str = 'r',
            encoding: Optional[str] = None,
            errors: Optional[str] = None,
            **kwargs: Any,
    ) -> typing.IO[AnyStr]:
        """Opens the file."""
    
    def read(self, mode='rb') -> bytes:
        with self.open(mode=mode) as f:
            return f.read()
    
    def readlines(self) -> List[str]:
        with self.open('r') as f:
            return f.readlines()

    def read_bytes(self) -> bytes:
        """Reads contents of self as bytes."""
        with self.open('rb') as f:
            return f.read()

    def read_text(self, encoding: Optional[str] = None) -> str:
        """Reads contents of self as bytes."""
        with self.open('r', encoding=encoding) as f:
            return f.read()

    def format(self: T, *args: Any, **kwargs: Any) -> T:
        """Apply `str.format()` to the path."""
        return type(self)(os.fspath(self).format(*args, **kwargs))    # pytype: disable=not-instantiable


class ReadWritePath(ReadOnlyPath, Protocol):
    """Protocol for pathlib.Path-like API.
    See [pathlib.Path](https://docs.python.org/3/library/pathlib.html)
    documentation.
    """

    @abc.abstractmethod
    def mkdir(
            self,
            mode: int = 0o777,
            parents: bool = False,
            exist_ok: bool = False,
    ) -> None:
        """Create a new directory at this given path."""

    @abc.abstractmethod
    def rmdir(self) -> None:
        """Remove the empty directory at this given path."""

    @abc.abstractmethod
    def rmtree(self) -> None:
        """Remove the directory, including all sub-files."""

    @abc.abstractmethod
    def unlink(self, missing_ok: bool = False) -> None:
        """Remove this file or symbolic link."""

    def write_bytes(self, data: bytes) -> None:
        """Writes content as bytes."""
        with self.open('wb') as f:
            return f.write(data)

    def write_text(
            self,
            data: str,
            encoding: Optional[str] = None,
            errors: Optional[str] = None,
    ) -> None:
        """Writes content as str."""
        with self.open('w') as f:
            return f.write(data)

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        """Create a file at this given path."""
        del mode    # Unused
        if self.exists():
            if exist_ok:
                return
            else:
                raise FileExistsError(f'{self} already exists.')
        self.write_text('')

    @abc.abstractmethod
    def rename(self: T, target: PathLike) -> T:
        """Renames the path."""

    @abc.abstractmethod
    def replace(self: T, target: PathLike) -> T:
        """Overwrites the destination path."""

    @abc.abstractmethod
    def copy(self: T, dst: PathLike, overwrite: bool = False) -> T:
        """Copy the current file to the given destination."""



"""
Adapted from https://github.com/tensorflow/datasets/blob/v4.4.0/tensorflow_datasets/core/utils/gpath.py
"""

import ntpath
import os
import pathlib
import posixpath
import types
import typing
from typing import Any, ClassVar, Iterator, Optional, Type, TypeVar, Union

from ..utils import lazy_import
from fileio.src import type_utils

tf = lazy_import('tensorflow')

_P = TypeVar('_P')


URI_PREFIXES = ('gs://', 's3://')
_URI_SCHEMES = frozenset(('gs', 's3'))

_URI_MAP_ROOT = {
        'gs://': '/gs/',
        's3://': '/s3/',
}


class _GPath(pathlib.PurePath, type_utils.ReadWritePath):
    """Pathlib like api around `tf.io.gfile`."""

    # `_PATH` is `posixpath` or `ntpath`.
    # Use explicit `join()` rather than `super().joinpath()` to avoid infinite
    # recursion.
    # Do not use `os.path`, so `PosixGPath('gs://abc')` works on windows.
    _PATH: ClassVar[types.ModuleType]

    def __new__(cls: Type[_P], *parts: type_utils.PathLike) -> _P:
        full_path = '/'.join(os.fspath(p) for p in parts)
        if not full_path.startswith(URI_PREFIXES):
            return super().__new__(cls, *parts)

        prefix = full_path[:5]
        new_prefix = _URI_MAP_ROOT[prefix]
        return super().__new__(cls, full_path.replace(prefix, new_prefix, 1))

    def _new(self: _P, *parts: type_utils.PathLike) -> _P:
        """Create a new `Path` child of same type."""
        return type(self)(*parts)

    # Could try to use `cached_property` when beam is compatible (currently
    # raise mutable input error).
    @property
    def _uri_scheme(self) -> Optional[str]:
        if (len(self.parts) >= 2 and self.parts[0] == '/' and
                self.parts[1] in _URI_SCHEMES):
            return self.parts[1]
        else:
            return None

    @property
    def _path_str(self) -> str:
        """Returns the `__fspath__` string representation."""
        uri_scheme = self._uri_scheme
        if uri_scheme:    # pylint: disable=using-constant-test
            return self._PATH.join(f'{uri_scheme}://', *self.parts[2:])
        else:
            return self._PATH.join(*self.parts) if self.parts else '.'

    def __fspath__(self) -> str:
        return self._path_str

    def __str__(self) -> str:    # pylint: disable=invalid-str-returned
        return self._path_str

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._path_str!r})'

    def exists(self) -> bool:
        """Returns True if self exists."""
        return tf.io.gfile.exists(self._path_str)

    def is_dir(self) -> bool:
        """Returns True if self is a directory."""
        return tf.io.gfile.isdir(self._path_str)

    def iterdir(self: _P) -> Iterator[_P]:
        """Iterates over the directory."""
        for f in tf.io.gfile.listdir(self._path_str):
            yield self._new(self, f)

    def expanduser(self: _P) -> _P:
        """Returns a new path with expanded `~` and `~user` constructs."""
        return self._new(self._PATH.expanduser(self._path_str))

    def resolve(self: _P, strict: bool = False) -> _P:
        """Returns the abolute path."""
        if self.is_cloud:
            return self._new(self.as_posix())
        return self._new(self._PATH.abspath(self._path_str))

    def glob(self: _P, pattern: str) -> Iterator[_P]:
        """Yielding all matching files (of any kind)."""
        for f in tf.io.gfile.glob(self._PATH.join(self._path_str, pattern)):
            yield self._new(f)

    def mkdir(
            self,
            mode: int = 0o777,
            parents: bool = False,
            exist_ok: bool = False,
    ) -> None:
        """Create a new directory at this given path."""
        if self.exists() and not exist_ok:
            raise FileExistsError(f'{self._path_str} already exists.')

        if parents:
            tf.io.gfile.makedirs(self._path_str)
        else:
            tf.io.gfile.mkdir(self._path_str)

    def rmdir(self) -> None:
        """Remove the empty directory."""
        if not self.is_dir():
            raise NotADirectoryError(f'{self._path_str} is not a directory.')
        if list(self.iterdir()):
            raise ValueError(f'Directory {self._path_str} is not empty')
        tf.io.gfile.rmtree(self._path_str)

    def rmtree(self) -> None:
        """Remove the directory."""
        tf.io.gfile.rmtree(self._path_str)

    def unlink(self, missing_ok: bool = False) -> None:
        """Remove this file or symbolic link."""
        try:
            tf.io.gfile.remove(self._path_str)
        except tf.errors.NotFoundError as e:
            if not missing_ok:
                raise FileNotFoundError(str(e))

    def open(
            self,
            mode: str = 'r',
            encoding: Optional[str] = None,
            errors: Optional[str] = None,
            **kwargs: Any,
    ) -> typing.IO[Union[str, bytes]]:
        """Opens the file."""
        if errors:
            raise NotImplementedError
        if encoding and not encoding.lower().startswith(('utf8', 'utf-8')):
            raise ValueError(f'Only UTF-8 encoding supported. Not: {encoding}')
        gfile = tf.io.gfile.GFile(self._path_str, mode, **kwargs)
        gfile = typing.cast(typing.IO[Union[str, bytes]], gfile)
        return gfile

    def rename(self: _P, target: type_utils.PathLike) -> _P:
        """Rename file or directory to the given target."""
        # Note: Issue if WindowsPath and target is gs://. Rather than using `_new`,
        # `GPath.__new__` should dynamically return either `PosixGPath` or
        # `WindowsPath`, similarly to `pathlib.Path`.
        target = self._new(target)
        tf.io.gfile.rename(self._path_str, os.fspath(target))
        return target

    def replace(self: _P, target: type_utils.PathLike) -> _P:
        """Replace file or directory to the given target."""
        target = self._new(target)
        tf.io.gfile.rename(self._path_str, os.fspath(target), overwrite=True)
        return target

    def copy(self: _P, dst: type_utils.PathLike, overwrite: bool = False, skip_errors=False) -> _P:
        """Copies the File to the Dir/File."""
        # Could add a recursive=True mode
        dst = self._new(dst) if isinstance(dst, str) else dst
        if not dst.is_file(): dst = dst.parent.joinpath(self.name)
        try:
            tf.io.gfile.copy(self._path_str, os.fspath(dst), overwrite=overwrite)
            return dst
        except Exception as e:
            print(f'Error Copying {self._path_str} -> {dst.as_posix()}: {str(e)}')
            if skip_errors:
                return None
            raise ValueError
        
    def copydir(self: _P, dst: type_utils.PathLike, ignore=['.git'], overwrite: bool = False, dryrun: bool = False):
        """Copies the Current Top Level Parent Dir to the Dst Dir without recursion"""
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        if not dryrun: dst.ensure_dir()
        copied_files = []
        fnames = self.listdir(ignore=ignore)
        curdir = self.absolute_parent
        for fname in fnames:
            dest_path = dst.joinpath(fname.relative_to(curdir))
            if not dryrun:
                fname.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files


    def copydirs(self: _P, dst: type_utils.PathLike, mode: str = 'shallow', pattern='*', ignore=['.git'], overwrite: bool = False, levels: int = 2, dryrun: bool = False):
        """Copies the Current Parent Dir to the Dst Dir.
        modes = [shallow for top level recursive. recursive for all nested]
        levels = number of recursive levels
        dryrun = returns all files that would have been copied without copying
        """
        assert mode in {'shallow', 'recursive'}, 'Invalid Mode Option: [shallow, recursive]'
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        levels = max(1, levels)
        dst.ensure_dir()
        curdir = self.absolute_parent
        copied_files = []
        if levels > 1 and mode == 'recursive' and '/' not in pattern:
            for _ in range(1, levels): pattern += '/*'
        if self.is_dir() and not pattern.startswith('/'): pattern = '*/' + pattern
        fiter = curdir.glob(pattern) if mode == 'shallow' else curdir.rglob(pattern)
        fnames = [f for f in fiter if not bool(set(f.parts).intersection(ignore))]
        print(len(fnames))
        for f in fnames:
            dest_path = dst.joinpath(f.relative_to(curdir))
            if not dryrun:
                if f.is_dir():
                    dest_path.ensure_dir()
                else:
                    f.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files

    def listdir(self: _P, ignore=['.git'], skip_dirs=True, skip_files=False):
        fnames = [f for f in self.iterdir() if not bool(set(f.parts).intersection(ignore))]
        fnames = [f.resolve() for f in fnames]
        if skip_dirs:
            return [f for f in fnames if f.is_file()]
        if skip_files:
            return [f for f in fnames if f.is_dir()]
        return fnames

    def listdirs(self: _P, mode: str = 'shallow', pattern='*', ignore=['.git'], skip_dirs=True, skip_files=False, levels: int = 2):
        """Lists all files in current parent dir
        modes = [shallow for top level recursive. recursive for all nested]
        """
        assert mode in {'shallow', 'recursive'}, 'Invalid Mode Option: [shallow, recursive]'
        curdir = self.absolute_parent
        levels = max(1, levels)
        if levels > 1 and mode == 'recursive' and '/' not in pattern:
            for _ in range(1, levels): pattern += '/*'
        if self.is_dir() and not pattern.startswith('*/'): pattern = '*/' + pattern
        print(pattern)
        fiter = curdir.glob(pattern) if mode == 'shallow' else curdir.rglob(pattern)
        fnames = [f for f in fiter if not bool(set(f.parts).intersection(ignore))]
        print(len(fnames))
        if skip_dirs:
            return [f for f in fnames if f.is_file()]
        if skip_files:
            return [f for f in fnames if f.is_dir()]
        return [f for f in fnames]


    def ensure_dir(self: _P, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """Ensures the parent directory exists, creates if not"""
        return self.absolute_parent.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    @property
    def absolute_parent(self: _P) -> _P:
        uri_scheme = self._uri_scheme
        if uri_scheme:
            return self._new(self._PATH.join(f'{uri_scheme}://', '/'.join(self.parts[2:-1])))
        p = self.resolve()
        if p.is_dir: 
            return p
        return p.parent

    @property
    def is_cloud(self: _P) -> bool:
        return bool(self._uri_scheme)
    
    @property
    def is_gcs(self: _P) -> bool:
        return bool(self._uri_scheme == 'gs')
    
    @property
    def is_s3(self: _P) -> bool:
        return bool(self._uri_scheme == 's3')
    
    @property
    def file_ext(self: _P) -> bool:
        return self.suffix[1:]
    


class PosixGPath(_GPath, pathlib.PurePosixPath):
    """Pathlib like api around `tf.io.gfile`."""
    _PATH = posixpath


class WindowsGPath(_GPath, pathlib.PureWindowsPath):
    """Pathlib like api around `tf.io.gfile`."""
    _PATH = ntpath


PathIOLike = TypeVar("PathIOLike", str, Union[pathlib.Path, PosixGPath])



"""
adapted from https://github.com/tensorflow/datasets/blob/v4.4.0/tensorflow_datasets/core/constants.py
"""


import os
from typing import List, Optional


from fileio.src import type_utils

# Github base URL
SRC_BASE_URL = 'https://github.com/tensorflow/datasets/tree/master/'

# Directory where to store processed datasets.
# If modifying this, should also update `scripts/cli/build.py` `--data_dir`
DATA_DIR = os.environ.get('FILEIO_DATA_DIR', os.path.join('~', 'fileio'))

# Suffix of files / directories which aren't finished downloading / extracting.
INCOMPLETE_SUFFIX = '.incomplete'

# Note: GCS constants are defined in `src/gcs_utils.py`

_registered_data_dir = set()


def add_data_dir(data_dir):
    """Registers a new default `data_dir` to search for datasets.
    When a `tfds.core.DatasetBuilder` is created with `data_dir=None`, TFDS
    will look in all registered `data_dir` (including the default one) to
    load existing datasets.
    * An error is raised if a dataset can be loaded from more than 1 registered
        data_dir.
    * This only affects reading datasets. Generation always uses the
        `data_dir` kwargs when specified or `tfds.core.constant.DATA_DIR` otherwise.
    Args:
        data_dir: New data_dir to register.
    """
    # Remove trailing / to avoid same directory being included twice in the set
    # with and without a final slash.
    data_dir = data_dir.rstrip('/')
    _registered_data_dir.add(data_dir)


def list_data_dirs(given_data_dir: Optional[str] = None,) -> List[str]:
    """Return the list of all `data_dir` to look-up.
    Args:
        given_data_dir: If a `data_dir` is provided, only the explicitly given
            `data_dir` will be returned, otherwise the list of all registered data_dir
            is returned
    Returns:
        The list of all data_dirs to look-up.
    """
    # If the data dir is explicitly given, no need to search everywhere.
    if given_data_dir:
        return [given_data_dir]
    else:
        all_data_dirs = _registered_data_dir | {DATA_DIR}
        return sorted(os.path.expanduser(d) for d in all_data_dirs)


def get_default_data_dir(given_data_dir: Optional[str] = None,) -> str:
    """Returns the default data_dir."""
    if given_data_dir:
        return os.path.expanduser(given_data_dir)
    else:
        return os.path.expanduser(DATA_DIR)


from .core import (
    File,
    gfile,
    torchdevice,
    _pickler,
    LineSeekableFile,
    timestamp,
    ftimestamp,
    glob,
    gcopy,
    isdir,
    isfile,
    listdir,
    mkdirs,
    mv,
    exists,
    rmdir,
    rm, 
    jparser,
    json,
    TextLineDataset,
    TFRecordDataset,
    TFRecordWriter,
    lazy_install,
    lazy_import,
    iterator_function,
    get_pathlike,
    filter_files,
    gsutil_exec,
    gsutil_rsync,
    gsutil_sync,

)

from .generic_path import as_path
from .gpath import PathIOLike
from .data_ops import autojson, read_json, read_jsonlines


PathIO = as_path
pickler = _pickler
device = torchdevice

__all__ = [
    "File",
    "gfile",
    "torchdevice",
    "device",
    "_pickler",
    "pickler",
    "LineSeekableFile",
    "timestamp",
    "ftimestamp",
    "as_path",
    "PathIOLike",
    "PathIO",
    "glob",
    "gcopy",
    "isdir",
    "isfile",
    "listdir",
    "mkdirs",
    "mv",
    "exists",
    "rmdir",
    "rm", 
    "jparser",
    "json",
    "TextLineDataset",
    "TFRecordDataset",
    "TFRecordWriter",
    "lazy_install",
    "lazy_import",
    "iterator_function",
    "get_pathlike",
    "filter_files",
    "gsutil_rsync",
    "gsutil_sync",
    "autojson", 
    "read_json", 
    "read_jsonlines"
]